# Copyright (c) MONAI Consortium
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#     http://www.apache.org/licenses/LICENSE-2.0
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import annotations

import sys
import unittest
from enum import Enum

import numpy as np
import torch
from parameterized import parameterized

from monai.data import CacheDataset, DataLoader, Dataset, create_test_image_2d
from monai.data.utils import decollate_batch
from monai.transforms import (
    AddChannel,
    AddChanneld,
    Compose,
    LoadImage,
    LoadImaged,
    RandAffine,
    RandFlip,
    RandFlipd,
    RandRotate90,
    SpatialPad,
    SpatialPadd,
    ToTensor,
    ToTensord,
)
from monai.transforms.inverse_batch_transform import Decollated
from monai.transforms.spatial.dictionary import RandAffined, RandRotate90d
from monai.utils import optional_import, set_determinism
from monai.utils.enums import PostFix, TraceKeys
from tests.utils import make_nifti_image

_, has_nib = optional_import("nibabel")

KEYS = ["image"]

TESTS_DICT: list[tuple] = []
TESTS_DICT.append((SpatialPadd(KEYS, 150), RandFlipd(KEYS, prob=1.0, spatial_axis=1)))
TESTS_DICT.append((RandRotate90d(KEYS, prob=0.0, max_k=1),))
TESTS_DICT.append((RandAffined(KEYS, prob=0.0, translate_range=10),))

TESTS_LIST: list[tuple] = []
TESTS_LIST.append((SpatialPad(150), RandFlip(prob=1.0, spatial_axis=1)))
TESTS_LIST.append((RandRotate90(prob=0.0, max_k=1),))
TESTS_LIST.append((RandAffine(prob=0.0, translate_range=10),))

TEST_BASIC = [
    [("channel", "channel"), ["channel", "channel"]],
    [torch.Tensor([1, 2, 3]), [torch.tensor(1.0), torch.tensor(2.0), torch.tensor(3.0)]],
    [
        [[torch.Tensor((1.0, 2.0, 3.0)), torch.Tensor((2.0, 3.0, 1.0))]],
        [
            [[torch.tensor(1.0), torch.tensor(2.0)]],
            [[torch.tensor(2.0), torch.tensor(3.0)]],
            [[torch.tensor(3.0), torch.tensor(1.0)]],
        ],
    ],
    [torch.Tensor((True, True, False, False)), [1.0, 1.0, 0.0, 0.0]],
    [
        [torch.Tensor([1]), torch.Tensor([2]), torch.Tensor([3])],
        [[torch.tensor(1.0), torch.tensor(2.0), torch.tensor(3.0)]],
    ],
    [[None, None], [None, None]],
    [["test"], ["test"]],
    [np.array([64, 64]), [64, 64]],
    [[], []],
    [[("ch1", "ch2"), ("ch3",)], [["ch1", "ch3"], ["ch2", None]]],  # default pad None
]


class TestDeCollate(unittest.TestCase):
    def setUp(self) -> None:
        set_determinism(seed=0)

        im = create_test_image_2d(100, 101)[0]
        self.data_dict = [{"image": make_nifti_image(im) if has_nib else im} for _ in range(6)]
        self.data_list = [make_nifti_image(im) if has_nib else im for _ in range(6)]

    def tearDown(self) -> None:
        set_determinism(None)

    def check_match(self, in1, in2):
        if isinstance(in1, dict):
            self.assertTrue(isinstance(in2, dict))
            for (k1, v1), (k2, v2) in zip(in1.items(), in2.items()):
                if isinstance(k1, Enum) and isinstance(k2, Enum):
                    k1, k2 = k1.value, k2.value
                self.check_match(k1, k2)
                # Transform ids won't match for windows with multiprocessing, so don't check values
                if k1 == TraceKeys.ID and sys.platform in ["darwin", "win32"]:
                    continue
                if not (isinstance(k1, str) and k1.endswith("_transforms")):
                    self.check_match(v1, v2)  # transform stack not necessarily match
        elif isinstance(in1, (list, tuple)):
            for l1, l2 in zip(in1, in2):
                self.check_match(l1, l2)
        elif isinstance(in1, (str, int)):
            self.assertEqual(in1, in2)
        elif isinstance(in1, (torch.Tensor, np.ndarray)):
            np.testing.assert_array_equal(in1, in2)
        else:
            raise RuntimeError(f"Not sure how to compare types. type(in1): {type(in1)}, type(in2): {type(in2)}")

    def check_decollate(self, dataset):
        batch_size = 2
        num_workers = 2 if sys.platform == "linux" else 0

        loader = DataLoader(dataset, batch_size=batch_size, shuffle=False, num_workers=num_workers)

        for b, batch_data in enumerate(loader):
            decollated_1 = decollate_batch(batch_data)
            decollated_2 = Decollated(detach=True)(batch_data)

            for decollated in [decollated_1, decollated_2]:
                for i, d in enumerate(decollated):
                    self.check_match(dataset[b * batch_size + i], d)

    @parameterized.expand(TESTS_DICT)
    def test_decollation_dict(self, *transforms):
        t_compose = Compose([AddChanneld(KEYS), Compose(transforms), ToTensord(KEYS)])
        # If nibabel present, read from disk
        if has_nib:
            t_compose = Compose([LoadImaged("image", image_only=True), t_compose])

        dataset = CacheDataset(self.data_dict, t_compose, progress=False)
        self.check_decollate(dataset=dataset)

    @parameterized.expand(TESTS_LIST)
    def test_decollation_tensor(self, *transforms):
        t_compose = Compose([AddChannel(), Compose(transforms), ToTensor()])
        # If nibabel present, read from disk
        if has_nib:
            t_compose = Compose([LoadImage(image_only=True), t_compose])

        dataset = Dataset(self.data_list, t_compose)
        self.check_decollate(dataset=dataset)

    @parameterized.expand(TESTS_LIST)
    def test_decollation_list(self, *transforms):
        t_compose = Compose([AddChannel(), Compose(transforms), ToTensor()])
        # If nibabel present, read from disk
        if has_nib:
            t_compose = Compose([LoadImage(image_only=True), t_compose])

        dataset = Dataset(self.data_list, t_compose)
        self.check_decollate(dataset=dataset)


class TestBasicDeCollate(unittest.TestCase):
    @parameterized.expand(TEST_BASIC)
    def test_decollation_examples(self, input_val, expected_out):
        out = decollate_batch(input_val)
        self.assertListEqual(expected_out, out)

    def test_dict_examples(self):
        test_case = {"meta": {"out": ["test", "test"]}, PostFix.meta("image"): {"scl_slope": torch.Tensor((0.0, 0.0))}}
        out = decollate_batch(test_case)
        self.assertEqual(out[0]["meta"]["out"], "test")
        self.assertEqual(out[0][PostFix.meta("image")]["scl_slope"], 0.0)

        test_case = [torch.ones((2, 1, 10, 10)), torch.ones((2, 3, 5, 5))]
        out = decollate_batch(test_case)
        self.assertTupleEqual(out[0][0].shape, (1, 10, 10))
        self.assertTupleEqual(out[0][1].shape, (3, 5, 5))

        test_case = torch.rand((2, 1, 10, 10))
        out = decollate_batch(test_case)
        self.assertTupleEqual(out[0].shape, (1, 10, 10))

        test_case = [torch.tensor(0), torch.tensor(0)]
        out = decollate_batch(test_case, detach=True)
        self.assertListEqual([0, 0], out)
        self.assertFalse(isinstance(out[0], torch.Tensor))

        test_case = {"a": [torch.tensor(0), torch.tensor(0)]}
        out = decollate_batch(test_case, detach=False)
        self.assertListEqual([{"a": torch.tensor(0)}, {"a": torch.tensor(0)}], out)
        self.assertTrue(isinstance(out[0]["a"], torch.Tensor))

        test_case = [torch.tensor(0), torch.tensor(0)]
        out = decollate_batch(test_case, detach=False)
        self.assertListEqual(test_case, out)

        test_case = {
            "image": torch.tensor([[[1, 2]], [[3, 4]]]),
            "label": torch.tensor([[[5, 6]], [[7, 8]]]),
            "pred": torch.tensor([[[9, 10]], [[11, 12]]]),
            "out": ["test"],
        }
        out = decollate_batch(test_case, detach=False)
        self.assertEqual(out[0]["out"], "test")

        test_case = {
            "image": torch.tensor([[[1, 2, 3]], [[3, 4, 5]]]),
            "label": torch.tensor([[[5]], [[7]]]),
            "out": ["test"],
        }
        out = decollate_batch(test_case, detach=False, pad=False)
        self.assertEqual(len(out), 1)  # no padding
        out = decollate_batch(test_case, detach=False, pad=True, fill_value=0)
        self.assertEqual(out[1]["out"], 0)  # verify padding fill_value

    def test_decollated(self):
        test_case = {
            "image": torch.tensor([[[1, 2]], [[3, 4]]]),
            "meta": {"out": ["test", "test"]},
            PostFix.meta("image"): {"scl_slope": torch.Tensor((0.0, 0.0))},
            "loss": 0.85,
        }
        transform = Decollated(keys=["meta", PostFix.meta("image")], detach=False)
        out = transform(test_case)
        self.assertFalse("loss" in out)
        self.assertEqual(out[0]["meta"]["out"], "test")
        self.assertEqual(out[0][PostFix.meta("image")]["scl_slope"], 0.0)
        self.assertTrue(isinstance(out[0][PostFix.meta("image")]["scl_slope"], torch.Tensor))
        # decollate all data with keys=None
        transform = Decollated(keys=None, detach=True)
        out = transform(test_case)
        self.assertEqual(out[1]["loss"], 0.85)
        self.assertEqual(out[0]["meta"]["out"], "test")
        self.assertEqual(out[0][PostFix.meta("image")]["scl_slope"], 0.0)
        self.assertTrue(isinstance(out[0][PostFix.meta("image")]["scl_slope"], float))

        # test list input
        test_case = [
            torch.tensor([[[1, 2]], [[3, 4]]]),
            {"out": ["test", "test"]},
            {"scl_slope": torch.Tensor((0.0, 0.0))},
            {"out2": ["test1"]},
            0.85,
            [],
        ]
        transform = Decollated(keys=None, detach=False, fill_value=-1)
        out = transform(test_case)

        self.assertEqual(out[0][-2], 0.85)  # scalar replicates
        self.assertEqual(out[1][-2], 0.85)  # scalar replicates
        self.assertEqual(out[1][-3], -1)  # fill value for the dictionary item
        self.assertEqual(out[0][1]["out"], "test")
        self.assertEqual(out[0][2]["scl_slope"], 0.0)
        self.assertTrue(isinstance(out[0][2]["scl_slope"], torch.Tensor))


if __name__ == "__main__":
    unittest.main()



# Copyright (c) MONAI Consortium
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#     http://www.apache.org/licenses/LICENSE-2.0
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import annotations

import unittest

import numpy as np
from parameterized import parameterized

from monai.transforms import GridDistortion
from tests.utils import TEST_NDARRAYS_ALL, assert_allclose

TESTS = []
for p in TEST_NDARRAYS_ALL:
    TESTS.append(
        [
            dict(num_cells=3, distort_steps=[(1.5,) * 4] * 2, mode="nearest", padding_mode="zeros"),
            p(np.indices([6, 6]).astype(np.float32)),
            p(
                np.array(
                    [
                        [
                            [0.0, 0.0, 0.0, 0.0, 0.0, 0.0],
                            [3.0, 3.0, 3.0, 0.0, 0.0, 0.0],
                            [3.0, 3.0, 3.0, 0.0, 0.0, 0.0],
                            [0.0, 0.0, 0.0, 0.0, 0.0, 0.0],
                            [0.0, 0.0, 0.0, 0.0, 0.0, 0.0],
                            [0.0, 0.0, 0.0, 0.0, 0.0, 0.0],
                        ],
                        [
                            [0.0, 3.0, 3.0, 0.0, 0.0, 0.0],
                            [0.0, 3.0, 3.0, 0.0, 0.0, 0.0],
                            [0.0, 3.0, 3.0, 0.0, 0.0, 0.0],
                            [0.0, 0.0, 0.0, 0.0, 0.0, 0.0],
                            [0.0, 0.0, 0.0, 0.0, 0.0, 0.0],
                            [0.0, 0.0, 0.0, 0.0, 0.0, 0.0],
                        ],
                    ]
                ).astype(np.float32)
            ),
        ]
    )
    num_cells = (2, 2)
    distort_steps = [(1.5,) * (1 + num_cells[0]), (1.0,) * (1 + num_cells[1])]
    TESTS.append(
        [
            dict(num_cells=num_cells, distort_steps=distort_steps, mode="bilinear", padding_mode="reflection"),
            p(np.indices([6, 6]).astype(np.float32)),
            p(
                np.array(
                    [
                        [
                            [0.0, 0.0, 0.0, 0.0, 0.0, 0.0],
                            [2.25, 2.25, 2.25, 2.25, 2.25, 2.25],
                            [4.5, 4.5, 4.5, 4.5, 4.5, 4.5],
                            [4.5, 4.5, 4.5, 4.5, 4.5, 4.5],
                            [4.2500, 4.2500, 4.2500, 4.2500, 4.2500, 4.2500],
                            [2.0, 2.0, 2.0, 2.0, 2.0, 2.0],
                        ],
                        [
                            [0.0, 1.5, 3.0, 3.0, 4.5, 5.0],
                            [0.0, 1.5, 3.0, 3.0, 4.5, 5.0],
                            [0.0, 1.5, 3.0, 3.0, 4.5, 5.0],
                            [0.0, 1.5, 3.0, 3.0, 4.5, 5.0],
                            [0.0, 1.5, 3.0, 3.0, 4.5, 5.0],
                            [0.0, 1.5, 3.0, 3.0, 4.5, 5.0],
                        ],
                    ]
                ).astype(np.float32)
            ),
        ]
    )
    TESTS.append(
        [
            dict(num_cells=2, distort_steps=[(1.25,) * 3] * 3, mode="nearest", padding_mode="zeros"),
            p(np.indices([3, 3, 3])[:1].astype(np.float32)),
            p(
                np.array(
                    [
                        [
                            [[0.0, 0.0, 0.0], [0.0, 0.0, 0.0], [0.0, 0.0, 0.0]],
                            [[1.0, 1.0, 0.0], [1.0, 1.0, 0.0], [0.0, 0.0, 0.0]],
                            [[0.0, 0.0, 0.0], [0.0, 0.0, 0.0], [0.0, 0.0, 0.0]],
                        ]
                    ]
                ).astype(np.float32)
            ),
        ]
    )


class TestGridDistortion(unittest.TestCase):
    @parameterized.expand(TESTS)
    def test_grid_distortion(self, input_param, input_data, expected_val):
        g = GridDistortion(**input_param)
        result = g(input_data)
        if input_param["padding_mode"] != "reflection":
            assert_allclose(result, expected_val, type_test=False, rtol=1e-4, atol=1e-4)
        else:
            assert_allclose(result.shape, expected_val.shape, type_test=False, rtol=1e-4, atol=1e-4)


if __name__ == "__main__":
    unittest.main()



# Copyright (c) MONAI Consortium
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#     http://www.apache.org/licenses/LICENSE-2.0
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import annotations

import unittest

import torch
from parameterized import parameterized

from monai.networks import eval_mode
from monai.networks.blocks.localnet_block import (
    LocalNetDownSampleBlock,
    LocalNetFeatureExtractorBlock,
    LocalNetUpSampleBlock,
)

TEST_CASE_DOWN_SAMPLE = [
    [{"spatial_dims": spatial_dims, "in_channels": 2, "out_channels": 4, "kernel_size": 3}] for spatial_dims in [2, 3]
]

TEST_CASE_UP_SAMPLE = [
    [
        {
            "spatial_dims": spatial_dims,
            "in_channels": 4,
            "out_channels": 2,
            "mode": "bilinear" if spatial_dims == 2 else "trilinear",
        }
    ]
    for spatial_dims in [2, 3]
]

TEST_CASE_EXTRACT = [
    [{"spatial_dims": spatial_dims, "in_channels": 2, "out_channels": 3, "act": act, "initializer": initializer}]
    for spatial_dims, act, initializer in zip([2, 3], ["sigmoid", None], ["kaiming_uniform", "zeros"])
]

in_size = 4


class TestLocalNetDownSampleBlock(unittest.TestCase):
    @parameterized.expand(TEST_CASE_DOWN_SAMPLE)
    def test_shape(self, input_param):
        net = LocalNetDownSampleBlock(**input_param)
        input_shape = (1, input_param["in_channels"], *([in_size] * input_param["spatial_dims"]))
        expect_mid_shape = (1, input_param["out_channels"], *([in_size] * input_param["spatial_dims"]))
        expect_x_shape = (1, input_param["out_channels"], *([in_size / 2] * input_param["spatial_dims"]))
        with eval_mode(net):
            x, mid = net(torch.randn(input_shape))
            self.assertEqual(x.shape, expect_x_shape)
            self.assertEqual(mid.shape, expect_mid_shape)

    def test_ill_arg(self):
        # even kernel_size
        with self.assertRaises(NotImplementedError):
            LocalNetDownSampleBlock(spatial_dims=2, in_channels=2, out_channels=4, kernel_size=4)

    @parameterized.expand(TEST_CASE_DOWN_SAMPLE)
    def test_ill_shape(self, input_param):
        net = LocalNetDownSampleBlock(**input_param)
        input_shape = (1, input_param["in_channels"], *([5] * input_param["spatial_dims"]))
        with self.assertRaises(ValueError):
            with eval_mode(net):
                net(torch.randn(input_shape))


class TestLocalNetUpSampleBlock(unittest.TestCase):
    @parameterized.expand(TEST_CASE_UP_SAMPLE)
    def test_shape(self, input_param):
        net = LocalNetUpSampleBlock(**input_param)
        input_shape = (1, input_param["in_channels"], *([in_size] * input_param["spatial_dims"]))
        mid_shape = (1, input_param["out_channels"], *([in_size * 2] * input_param["spatial_dims"]))
        expected_shape = mid_shape
        with eval_mode(net):
            result = net(torch.randn(input_shape), torch.randn(mid_shape))
            self.assertEqual(result.shape, expected_shape)

    def test_ill_arg(self):
        # channel unmatch
        with self.assertRaises(ValueError):
            LocalNetUpSampleBlock(spatial_dims=2, in_channels=2, out_channels=2)

    @parameterized.expand(TEST_CASE_UP_SAMPLE)
    def test_ill_shape(self, input_param):
        net = LocalNetUpSampleBlock(**input_param)
        input_shape = (1, input_param["in_channels"], *([in_size] * input_param["spatial_dims"]))
        mid_shape = (1, input_param["out_channels"], *([in_size] * input_param["spatial_dims"]))
        with self.assertRaises(ValueError):
            with eval_mode(net):
                net(torch.randn(input_shape), torch.randn(mid_shape))


class TestExtractBlock(unittest.TestCase):
    @parameterized.expand(TEST_CASE_EXTRACT)
    def test_shape(self, input_param):
        net = LocalNetFeatureExtractorBlock(**input_param)
        input_shape = (1, input_param["in_channels"], *([in_size] * input_param["spatial_dims"]))
        expected_shape = (1, input_param["out_channels"], *([in_size] * input_param["spatial_dims"]))
        with eval_mode(net):
            result = net(torch.randn(input_shape))
            self.assertEqual(result.shape, expected_shape)

    def test_ill_arg(self):
        with self.assertRaises(ValueError):
            LocalNetFeatureExtractorBlock(spatial_dims=2, in_channels=2, out_channels=2, initializer="none")


if __name__ == "__main__":
    unittest.main()



# Copyright (c) MONAI Consortium
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#     http://www.apache.org/licenses/LICENSE-2.0
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import annotations

import unittest

import numpy as np
from parameterized import parameterized

from monai.data import MetaTensor, set_track_meta
from monai.transforms import Affine, Rotate90
from monai.transforms.lazy.functional import apply_transforms
from monai.utils import optional_import
from tests.lazy_transforms_utils import test_resampler_lazy
from tests.utils import (
    TEST_NDARRAYS_ALL,
    NumpyImageTestCase2D,
    NumpyImageTestCase3D,
    assert_allclose,
    test_local_inversion,
)


class TestRotate90(NumpyImageTestCase2D):
    def test_rotate90_default(self):
        rotate = Rotate90()
        for p in TEST_NDARRAYS_ALL:
            im = p(self.imt[0])
            set_track_meta(True)
            call_param = {"img": im}
            rotated = rotate(**call_param)

            # test lazy
            test_resampler_lazy(rotate, rotated, call_param=call_param)
            rotate.lazy_evaluation = False

            test_local_inversion(rotate, rotated, im)
            expected = [np.rot90(channel, 1, (0, 1)) for channel in self.imt[0]]
            expected = np.stack(expected)
            assert_allclose(rotated, p(expected), rtol=1.0e-5, atol=1.0e-8, type_test="tensor")
            set_track_meta(False)
            rotated = rotate(im)
            self.assertNotIsInstance(rotated, MetaTensor)
            set_track_meta(True)

    def test_k(self):
        rotate = Rotate90(k=2)
        for p in TEST_NDARRAYS_ALL:
            im = p(self.imt[0])
            call_param = {"img": im}
            rotated = rotate(**call_param)

            # test lazy
            test_resampler_lazy(rotate, rotated, call_param=call_param)
            rotate.lazy_evaluation = False

            test_local_inversion(rotate, rotated, im)
            expected = [np.rot90(channel, 2, (0, 1)) for channel in self.imt[0]]
            expected = np.stack(expected)
            assert_allclose(rotated, p(expected), rtol=1.0e-5, atol=1.0e-8, type_test="tensor")

    def test_spatial_axes(self):
        rotate = Rotate90(spatial_axes=(0, -1))
        for p in TEST_NDARRAYS_ALL:
            im = p(self.imt[0])
            call_param = {"img": im}
            rotated = rotate(**call_param)

            # test lazy
            test_resampler_lazy(rotate, rotated, call_param=call_param)
            rotate.lazy_evaluation = False

            test_local_inversion(rotate, rotated, im)
            expected = [np.rot90(channel, 1, (0, -1)) for channel in self.imt[0]]
            expected = np.stack(expected)
            assert_allclose(rotated, p(expected), rtol=1.0e-5, atol=1.0e-8, type_test="tensor")

    def test_prob_k_spatial_axes(self):
        rotate = Rotate90(k=2, spatial_axes=(0, 1))
        for p in TEST_NDARRAYS_ALL:
            im = p(self.imt[0])
            call_param = {"img": im}
            rotated = rotate(**call_param)

            # test lazy
            test_resampler_lazy(rotate, rotated, call_param=call_param)
            rotate.lazy_evaluation = False

            test_local_inversion(rotate, rotated, im)
            expected = [np.rot90(channel, 2, (0, 1)) for channel in self.imt[0]]
            expected = np.stack(expected)
            assert_allclose(rotated, p(expected), rtol=1.0e-5, atol=1.0e-8, type_test="tensor")


class TestRotate903d(NumpyImageTestCase3D):
    def test_rotate90_default(self):
        rotate = Rotate90()
        for p in TEST_NDARRAYS_ALL:
            im = p(self.imt[0])
            call_param = {"img": im}
            rotated = rotate(**call_param)

            # test lazy
            test_resampler_lazy(rotate, rotated, call_param=call_param)
            rotate.lazy_evaluation = False

            test_local_inversion(rotate, rotated, im)
            expected = [np.rot90(channel, 1, (0, 1)) for channel in self.imt[0]]
            expected = np.stack(expected)
            assert_allclose(rotated, p(expected), rtol=1.0e-5, atol=1.0e-8, type_test="tensor")

    def test_k(self):
        rotate = Rotate90(k=2)
        for p in TEST_NDARRAYS_ALL:
            im = p(self.imt[0])
            call_param = {"img": im}
            rotated = rotate(**call_param)

            # test lazy
            test_resampler_lazy(rotate, rotated, call_param=call_param)
            rotate.lazy_evaluation = False

            test_local_inversion(rotate, rotated, im)
            expected = [np.rot90(channel, 2, (0, 1)) for channel in self.imt[0]]
            expected = np.stack(expected)
            assert_allclose(rotated, p(expected), rtol=1.0e-5, atol=1.0e-8, type_test="tensor")

    def test_spatial_axes(self):
        rotate = Rotate90(spatial_axes=(0, -1))
        for p in TEST_NDARRAYS_ALL:
            im = p(self.imt[0])
            call_param = {"img": im}
            rotated = rotate(**call_param)

            # test lazy
            test_resampler_lazy(rotate, rotated, call_param=call_param)
            rotate.lazy_evaluation = False

            test_local_inversion(rotate, rotated, im)
            expected = [np.rot90(channel, 1, (0, -1)) for channel in self.imt[0]]
            expected = np.stack(expected)
            assert_allclose(rotated, p(expected), rtol=1.0e-5, atol=1.0e-8, type_test="tensor")

    def test_prob_k_spatial_axes(self):
        rotate = Rotate90(k=2, spatial_axes=(0, 1))
        for p in TEST_NDARRAYS_ALL:
            im = p(self.imt[0])
            call_param = {"img": im}
            rotated = rotate(**call_param)

            # test lazy
            test_resampler_lazy(rotate, rotated, call_param=call_param)
            rotate.lazy_evaluation = False

            test_local_inversion(rotate, rotated, im)
            expected = [np.rot90(channel, 2, (0, 1)) for channel in self.imt[0]]
            expected = np.stack(expected)
            assert_allclose(rotated, p(expected), rtol=1.0e-5, atol=1.0e-8, type_test="tensor")


@unittest.skipUnless(optional_import("scipy")[1], "Requires scipy library.")
class TestRot90Consistency(unittest.TestCase):
    @parameterized.expand([[2], [3], [4]])
    def test_affine_rot90(self, s):
        """s"""
        im = np.arange(int(s * s)).reshape(1, s, s).astype(float)
        mat = np.array([[0, 1, 0], [-1, 0, 0], [0, 0, 1]])

        def method_0(im, ac):
            xform = Affine(align_corners=ac, affine=mat, image_only=True, spatial_size=s)
            xform.lazy_evaluation = True
            out = xform(im)
            out = apply_transforms(out, padding_mode="border", align_corners=ac)[0]
            return out

        def method_1(im, ac):
            xform = Affine(align_corners=ac, affine=mat, image_only=True, spatial_size=s)
            xform.lazy_evaluation = True
            out = xform(im)
            out = apply_transforms(out, mode=1, padding_mode="nearest", align_corners=ac)[0]
            return out

        def method_2(im, ac):
            xform = Affine(align_corners=ac, affine=mat, padding_mode="border", image_only=True, spatial_size=s)
            out = xform(im)
            return out

        def method_3(im, ac):
            xform = Affine(
                align_corners=ac, affine=mat, mode=1, padding_mode="nearest", image_only=True, spatial_size=s
            )
            out = xform(im)
            return out

        for call in (method_0, method_1, method_2, method_3):
            for ac in (False, True):
                out = call(im, ac)
                ref = Rotate90()(im)
                assert_allclose(out, ref, rtol=1e-4, atol=1e-4, type_test=False)


if __name__ == "__main__":
    unittest.main()



# Copyright (c) MONAI Consortium
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#     http://www.apache.org/licenses/LICENSE-2.0
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import annotations

import os
import tempfile
import unittest

import nibabel as nib
import numpy as np
from parameterized import parameterized

from monai.data import NibabelWriter
from monai.transforms import LoadImage, Orientation, Spacing
from tests.utils import TEST_NDARRAYS, assert_allclose, make_nifti_image

TESTS = []
for p in TEST_NDARRAYS:
    for q in TEST_NDARRAYS:
        TEST_IMAGE = p(np.arange(24).reshape((2, 4, 3)))
        TEST_AFFINE = q(
            np.array(
                [[-5.3, 0.0, 0.0, 102.01], [0.0, 0.52, 2.17, -7.50], [-0.0, 1.98, -0.26, -23.12], [0.0, 0.0, 0.0, 1.0]]
            )
        )
        TESTS.append(
            [
                TEST_IMAGE,
                TEST_AFFINE,
                dict(reader="NibabelReader", image_only=True, as_closest_canonical=True),
                np.array(
                    [
                        [[12.0, 15.0, 18.0, 21.0], [13.0, 16.0, 19.0, 22.0], [14.0, 17.0, 20.0, 23.0]],
                        [[0.0, 3.0, 6.0, 9.0], [1.0, 4.0, 7.0, 10.0], [2.0, 5.0, 8.0, 11.0]],
                    ]
                ),
            ]
        )
        TESTS.append(
            [
                TEST_IMAGE,
                TEST_AFFINE,
                dict(reader="NibabelReader", image_only=True, as_closest_canonical=False),
                np.arange(24).reshape((2, 4, 3)),
            ]
        )
        TESTS.append(
            [
                TEST_IMAGE,
                TEST_AFFINE,
                dict(reader="NibabelReader", image_only=True, as_closest_canonical=False),
                np.arange(24).reshape((2, 4, 3)),
            ]
        )
        TESTS.append(
            [
                TEST_IMAGE,
                None,
                dict(reader="NibabelReader", image_only=True, as_closest_canonical=False),
                np.arange(24).reshape((2, 4, 3)),
            ]
        )


class TestNiftiLoadRead(unittest.TestCase):
    @parameterized.expand(TESTS)
    def test_orientation(self, array, affine, reader_param, expected):
        test_image = make_nifti_image(array, affine)

        # read test cases
        loader = LoadImage(**reader_param)
        load_result = loader(test_image)
        data_array = load_result.numpy()
        if reader_param.get("image_only", False):
            header = None
        else:
            header = load_result.meta
            header["affine"] = header["affine"].numpy()
        if os.path.exists(test_image):
            os.remove(test_image)

        # write test cases
        writer_obj = NibabelWriter()
        writer_obj.set_data_array(data_array, channel_dim=None)
        if header is not None:
            writer_obj.set_metadata(header)
        elif affine is not None:
            writer_obj.set_metadata({"affine": affine})
        writer_obj.write(test_image, verbose=True)
        saved = nib.load(test_image)
        saved_affine = saved.affine
        saved_data = saved.get_fdata()
        if os.path.exists(test_image):
            os.remove(test_image)

        if affine is not None:
            assert_allclose(saved_affine, affine, type_test=False)
        assert_allclose(saved_data, expected, type_test=False)

    def test_consistency(self):
        np.set_printoptions(suppress=True, precision=3)
        test_image = make_nifti_image(np.arange(64).reshape(1, 8, 8), np.diag([1.5, 1.5, 1.5, 1]))
        data = LoadImage(image_only=True, reader="NibabelReader", as_closest_canonical=False)(test_image)
        header = data.meta
        data = Spacing([0.8, 0.8, 0.8])(data[None], header["affine"], mode="nearest")
        original_affine = data.meta["original_affine"]
        data = Orientation("ILP")(data)
        new_affine = data.affine
        if os.path.exists(test_image):
            os.remove(test_image)
        writer_obj = NibabelWriter()
        writer_obj.set_data_array(data[0], channel_dim=None)
        writer_obj.set_metadata(
            meta_dict={"affine": new_affine, "original_affine": original_affine}, mode="nearest", padding_mode="border"
        )
        writer_obj.write(test_image, verbose=True)
        saved = nib.load(test_image)
        saved_data = saved.get_fdata()
        np.testing.assert_allclose(saved_data, np.arange(64).reshape(1, 8, 8), atol=1e-7)
        if os.path.exists(test_image):
            os.remove(test_image)
        writer_obj.set_data_array(data[0], channel_dim=None)
        writer_obj.set_metadata(
            meta_dict={"affine": new_affine, "original_affine": original_affine, "spatial_shape": (1, 8, 8)},
            mode="nearest",
            padding_mode="border",
        )
        writer_obj.write(test_image, verbose=True)
        saved = nib.load(test_image)
        saved_data = saved.get_fdata()
        np.testing.assert_allclose(saved_data, np.arange(64).reshape(1, 8, 8), atol=1e-7)
        if os.path.exists(test_image):
            os.remove(test_image)
        # test the case no resample
        writer_obj.set_data_array(data[0], channel_dim=None)
        writer_obj.set_metadata(meta_dict={"affine": new_affine, "original_affine": original_affine}, resample=False)
        writer_obj.write(test_image, verbose=True)
        saved = nib.load(test_image)
        np.testing.assert_allclose(saved.affine, new_affine)
        if os.path.exists(test_image):
            os.remove(test_image)

    def test_write_2d(self):
        with tempfile.TemporaryDirectory() as out_dir:
            image_name = os.path.join(out_dir, "test.nii.gz")
            for p in TEST_NDARRAYS:
                img = p(np.arange(6).reshape((2, 3)))
                writer_obj = NibabelWriter()
                writer_obj.set_data_array(img, channel_dim=None)
                writer_obj.set_metadata({"affine": np.diag([1, 1, 1]), "original_affine": np.diag([1.4, 1, 1])})
                writer_obj.write(image_name, verbose=True)
                out = nib.load(image_name)
                np.testing.assert_allclose(out.get_fdata(), [[0, 1, 2], [3.0, 4, 5]], atol=1e-4, rtol=1e-4)
                np.testing.assert_allclose(out.affine, np.diag([1.4, 1, 1, 1]), atol=1e-4, rtol=1e-4)

                image_name = os.path.join(out_dir, "test1.nii.gz")
                img = np.arange(5).reshape((1, 5))
                writer_obj.set_data_array(img, channel_dim=None)
                writer_obj.set_metadata(
                    {"affine": np.diag([1, 1, 1, 3, 3]), "original_affine": np.diag([1.4, 2.0, 1, 3, 5])}
                )
                writer_obj.write(image_name, verbose=True)
                out = nib.load(image_name)
                np.testing.assert_allclose(out.get_fdata(), [[0, 2, 4]], atol=1e-4, rtol=1e-4)
                np.testing.assert_allclose(out.affine, np.diag([1.4, 2, 1, 1]), atol=1e-4, rtol=1e-4)

    def test_write_3d(self):
        with tempfile.TemporaryDirectory() as out_dir:
            image_name = os.path.join(out_dir, "test.nii.gz")
            for p in TEST_NDARRAYS:
                img = p(np.arange(6).reshape((1, 2, 3)))
                writer_obj = NibabelWriter()
                writer_obj.set_data_array(img, channel_dim=None)
                writer_obj.set_metadata({"affine": np.diag([1, 1, 1, 1]), "original_affine": np.diag([1.4, 1, 1, 1])})
                writer_obj.write(image_name, verbose=True)
                out = nib.load(image_name)
                np.testing.assert_allclose(out.get_fdata(), [[[0, 1, 2], [3, 4, 5]]], atol=1e-4, rtol=1e-4)
                np.testing.assert_allclose(out.affine, np.diag([1.4, 1, 1, 1]), atol=1e-4, rtol=1e-4)

                image_name = os.path.join(out_dir, "test1.nii.gz")
                img = p(np.arange(5).reshape((1, 1, 5)))
                writer_obj.set_data_array(img, channel_dim=None)
                writer_obj.set_metadata(
                    {"affine": np.diag([1, 1, 1, 3, 3]), "original_affine": np.diag([1.4, 2.0, 2, 3, 5])}
                )
                writer_obj.write(image_name, verbose=True)
                out = nib.load(image_name)
                np.testing.assert_allclose(out.get_fdata(), [[[0, 2, 4]]], atol=1e-4, rtol=1e-4)
                np.testing.assert_allclose(out.affine, np.diag([1.4, 2, 2, 1]), atol=1e-4, rtol=1e-4)

    def test_write_4d(self):
        with tempfile.TemporaryDirectory() as out_dir:
            image_name = os.path.join(out_dir, "test.nii.gz")
            for p in TEST_NDARRAYS:
                img = p(np.arange(6).reshape((1, 1, 3, 2)))
                writer_obj = NibabelWriter()
                writer_obj.set_data_array(img, channel_dim=-1)
                writer_obj.set_metadata({"affine": np.diag([1.4, 1, 1, 1]), "original_affine": np.diag([1, 1.4, 1, 1])})
                writer_obj.write(image_name, verbose=True)
                out = nib.load(image_name)
                np.testing.assert_allclose(out.get_fdata(), [[[[0, 1], [2, 3], [4, 5]]]], atol=1e-4, rtol=1e-4)
                np.testing.assert_allclose(out.affine, np.diag([1, 1.4, 1, 1]), atol=1e-4, rtol=1e-4)

                image_name = os.path.join(out_dir, "test1.nii.gz")
                img = p(np.arange(5).reshape((1, 1, 5, 1)))
                writer_obj.set_data_array(img, channel_dim=-1, squeeze_end_dims=False)
                writer_obj.set_metadata(
                    {"affine": np.diag([1, 1, 1, 3, 3]), "original_affine": np.diag([1.4, 2.0, 2, 3, 5])}
                )
                writer_obj.write(image_name, verbose=True)
                out = nib.load(image_name)
                np.testing.assert_allclose(out.get_fdata(), [[[[0], [2], [4]]]], atol=1e-4, rtol=1e-4)
                np.testing.assert_allclose(out.affine, np.diag([1.4, 2, 2, 1]), atol=1e-4, rtol=1e-4)

    def test_write_5d(self):
        with tempfile.TemporaryDirectory() as out_dir:
            image_name = os.path.join(out_dir, "test.nii.gz")
            for p in TEST_NDARRAYS:
                img = p(np.arange(12).reshape((1, 1, 3, 2, 2)))
                writer_obj = NibabelWriter()
                writer_obj.set_data_array(img, channel_dim=-1, squeeze_end_dims=False, spatial_ndim=None)
                writer_obj.set_metadata({"affine": np.diag([1, 1, 1, 1]), "original_affine": np.diag([1.4, 1, 1, 1])})
                writer_obj.write(image_name, verbose=True)
                out = nib.load(image_name)
                np.testing.assert_allclose(
                    out.get_fdata(),
                    np.array([[[[[0.0, 1.0], [2.0, 3.0]], [[4.0, 5.0], [6.0, 7.0]], [[8.0, 9.0], [10.0, 11.0]]]]]),
                    atol=1e-4,
                    rtol=1e-4,
                )
                np.testing.assert_allclose(out.affine, np.diag([1.4, 1, 1, 1]), atol=1e-4, rtol=1e-4)

                image_name = os.path.join(out_dir, "test1.nii.gz")
                img = p(np.arange(10).reshape((1, 1, 5, 1, 2)))
                writer_obj.set_data_array(img, channel_dim=-1, squeeze_end_dims=False, spatial_ndim=None)
                writer_obj.set_metadata({"affine": np.diag([1, 1, 1, 3]), "original_affine": np.diag([1.4, 2.0, 2, 3])})
                writer_obj.write(image_name, verbose=True)
                out = nib.load(image_name)
                np.testing.assert_allclose(
                    out.get_fdata(), np.array([[[[[0.0, 2.0]], [[4.0, 5.0]], [[7.0, 9.0]]]]]), atol=1e-4, rtol=1e-4
                )
                np.testing.assert_allclose(out.affine, np.diag([1.4, 2, 2, 1]), atol=1e-4, rtol=1e-4)


if __name__ == "__main__":
    unittest.main()



# Copyright (c) MONAI Consortium
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#     http://www.apache.org/licenses/LICENSE-2.0
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import annotations

import unittest

import torch
from parameterized import parameterized

from monai.networks.layers import ApplyFilter, EllipticalFilter, LaplaceFilter, MeanFilter, SharpenFilter

TEST_CASES_MEAN = [(3, 3, torch.ones(3, 3, 3)), (2, 5, torch.ones(5, 5))]

TEST_CASES_LAPLACE = [
    (
        3,
        3,
        torch.Tensor(
            [
                [[-1, -1, -1], [-1, -1, -1], [-1, -1, -1]],
                [[-1, -1, -1], [-1, 26, -1], [-1, -1, -1]],
                [[-1, -1, -1], [-1, -1, -1], [-1, -1, -1]],
            ]
        ),
    ),
    (2, 3, torch.Tensor([[-1, -1, -1], [-1, 8, -1], [-1, -1, -1]])),
]

TEST_CASES_ELLIPTICAL = [
    (
        3,
        3,
        torch.Tensor(
            [[[0, 0, 0], [0, 1, 0], [0, 0, 0]], [[0, 1, 0], [1, 1, 1], [0, 1, 0]], [[0, 0, 0], [0, 1, 0], [0, 0, 0]]]
        ),
    ),
    (2, 3, torch.Tensor([[0, 1, 0], [1, 1, 1], [0, 1, 0]])),
]

TEST_CASES_SHARPEN = [
    (
        3,
        3,
        torch.Tensor(
            [
                [[0, 0, 0], [0, -1, 0], [0, 0, 0]],
                [[0, -1, 0], [-1, 7, -1], [0, -1, 0]],
                [[0, 0, 0], [0, -1, 0], [0, 0, 0]],
            ]
        ),
    ),
    (2, 3, torch.Tensor([[0, -1, 0], [-1, 5, -1], [0, -1, 0]])),
]


class _TestFilter:
    def test_init(self, spatial_dims, size, expected):
        test_filter = self.filter_class(spatial_dims=spatial_dims, size=size)
        torch.testing.assert_allclose(expected, test_filter.filter)
        self.assertIsInstance(test_filter, torch.nn.Module)

    def test_forward(self):
        test_filter = self.filter_class(spatial_dims=2, size=3)
        input = torch.ones(1, 1, 5, 5)
        _ = test_filter(input)


class TestApplyFilter(unittest.TestCase):
    def test_init_and_forward_2d(self):
        filter_2d = torch.ones(3, 3)
        image_2d = torch.ones(1, 3, 3)
        apply_filter_2d = ApplyFilter(filter_2d)
        out = apply_filter_2d(image_2d)
        self.assertEqual(out.shape, image_2d.shape)

    def test_init_and_forward_3d(self):
        filter_2d = torch.ones(3, 3, 3)
        image_2d = torch.ones(1, 3, 3, 3)
        apply_filter_2d = ApplyFilter(filter_2d)
        out = apply_filter_2d(image_2d)
        self.assertEqual(out.shape, image_2d.shape)


class MeanFilterTestCase(_TestFilter, unittest.TestCase):
    def setUp(self) -> None:
        self.filter_class = MeanFilter

    @parameterized.expand(TEST_CASES_MEAN)
    def test_init(self, spatial_dims, size, expected):
        super().test_init(spatial_dims, size, expected)


class LaplaceFilterTestCase(_TestFilter, unittest.TestCase):
    def setUp(self) -> None:
        self.filter_class = LaplaceFilter

    @parameterized.expand(TEST_CASES_LAPLACE)
    def test_init(self, spatial_dims, size, expected):
        super().test_init(spatial_dims, size, expected)


class EllipticalTestCase(_TestFilter, unittest.TestCase):
    def setUp(self) -> None:
        self.filter_class = EllipticalFilter

    @parameterized.expand(TEST_CASES_ELLIPTICAL)
    def test_init(self, spatial_dims, size, expected):
        super().test_init(spatial_dims, size, expected)


class SharpenTestCase(_TestFilter, unittest.TestCase):
    def setUp(self) -> None:
        self.filter_class = SharpenFilter

    @parameterized.expand(TEST_CASES_SHARPEN)
    def test_init(self, spatial_dims, size, expected):
        super().test_init(spatial_dims, size, expected)


if __name__ == "__main__":
    unittest.main()



# Copyright (c) MONAI Consortium
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#     http://www.apache.org/licenses/LICENSE-2.0
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import annotations

import os
import unittest

from parameterized import parameterized

from monai.bundle import ConfigWorkflow
from monai.fl.client import MonaiAlgoStats
from monai.fl.utils.constants import ExtraItems, FlStatistics
from monai.fl.utils.exchange_object import ExchangeObject
from tests.utils import SkipIfNoModule

_root_dir = os.path.abspath(os.path.join(os.path.dirname(__file__)))
_data_dir = os.path.join(_root_dir, "testing_data")
_logging_file = os.path.join(_data_dir, "logging.conf")

TEST_GET_DATA_STATS_1 = [
    {
        "bundle_root": _data_dir,
        "workflow": ConfigWorkflow(
            workflow="train",
            config_file=os.path.join(_data_dir, "config_fl_stats_1.json"),
            logging_file=_logging_file,
            meta_file=None,
        ),
        "config_filters_filename": os.path.join(_data_dir, "config_fl_filters.json"),
    }
]
TEST_GET_DATA_STATS_2 = [
    {
        "bundle_root": _data_dir,
        "config_train_filename": os.path.join(_data_dir, "config_fl_stats_2.json"),
        "config_filters_filename": os.path.join(_data_dir, "config_fl_filters.json"),
    }
]
TEST_GET_DATA_STATS_3 = [
    {
        "bundle_root": _data_dir,
        "workflow": ConfigWorkflow(
            workflow="train",
            config_file=[
                os.path.join(_data_dir, "config_fl_stats_1.json"),
                os.path.join(_data_dir, "config_fl_stats_2.json"),
            ],
            logging_file=_logging_file,
            meta_file=None,
        ),
        "config_filters_filename": os.path.join(_data_dir, "config_fl_filters.json"),
    }
]


@SkipIfNoModule("ignite")
class TestFLMonaiAlgo(unittest.TestCase):
    @parameterized.expand([TEST_GET_DATA_STATS_1, TEST_GET_DATA_STATS_2, TEST_GET_DATA_STATS_3])
    def test_get_data_stats(self, input_params):
        # initialize algo
        algo = MonaiAlgoStats(**input_params)
        algo.initialize(extra={ExtraItems.CLIENT_NAME: "test_fl", ExtraItems.APP_ROOT: _data_dir})

        requested_stats = {FlStatistics.HIST_BINS: 100, FlStatistics.HIST_RANGE: [-500, 500]}
        # test train
        stats = algo.get_data_stats(extra=requested_stats)
        self.assertIsInstance(stats, ExchangeObject)


if __name__ == "__main__":
    unittest.main()



# Copyright (c) MONAI Consortium
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#     http://www.apache.org/licenses/LICENSE-2.0
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import annotations

import unittest

from parameterized import parameterized

from monai.transforms import MapTransform

TEST_CASES = [["item", ("item",)], [None, (None,)], [["item1", "item2"], ("item1", "item2")]]

TEST_ILL_CASES = [[ValueError, []], [ValueError, ()], [TypeError, [[]]]]


class MapTest(MapTransform):
    def __call__(self, data):
        pass


class TestRandomizable(unittest.TestCase):
    @parameterized.expand(TEST_CASES)
    def test_keys(self, keys, expected):
        transform = MapTest(keys=keys)
        self.assertEqual(transform.keys, expected)

    @parameterized.expand(TEST_ILL_CASES)
    def test_wrong_keys(self, exception, keys):
        with self.assertRaisesRegex(exception, ""):
            MapTest(keys=keys)


if __name__ == "__main__":
    unittest.main()



# Copyright (c) MONAI Consortium
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#     http://www.apache.org/licenses/LICENSE-2.0
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import annotations

import unittest
from typing import TYPE_CHECKING
from unittest import skipUnless

import numpy as np
from parameterized import parameterized

from monai.transforms import ToPIL
from monai.utils import optional_import
from tests.utils import TEST_NDARRAYS, assert_allclose

if TYPE_CHECKING:
    from PIL.Image import Image as PILImageImage
    from PIL.Image import fromarray as pil_image_fromarray

    has_pil = True
else:
    pil_image_fromarray, has_pil = optional_import("PIL.Image", name="fromarray")
    PILImageImage, _ = optional_import("PIL.Image", name="Image")

im = [[1.0, 2.0], [3.0, 4.0]]
TESTS = []
for p in TEST_NDARRAYS:
    TESTS.append([p(im)])
if has_pil:
    TESTS.append([pil_image_fromarray(np.array(im))])


class TestToPIL(unittest.TestCase):
    @parameterized.expand(TESTS)
    @skipUnless(has_pil, "Requires `pillow` package.")
    def test_value(self, test_data):
        result = ToPIL()(test_data)
        self.assertTrue(isinstance(result, PILImageImage))
        assert_allclose(np.array(result), test_data, type_test=False)


if __name__ == "__main__":
    unittest.main()



# Copyright (c) MONAI Consortium
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#     http://www.apache.org/licenses/LICENSE-2.0
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import annotations

import itertools
import os
import shutil
import tempfile
import unittest

import numpy as np
import torch
from parameterized import parameterized

from monai.data.image_reader import ITKReader, NibabelReader, NrrdReader, PILReader
from monai.data.image_writer import ITKWriter, NibabelWriter, PILWriter, register_writer, resolve_writer
from monai.data.meta_tensor import MetaTensor
from monai.transforms import LoadImage, SaveImage, moveaxis
from monai.utils import MetaKeys, OptionalImportError, optional_import
from tests.utils import TEST_NDARRAYS, assert_allclose

_, has_itk = optional_import("itk", allow_namespace_pkg=True)


@unittest.skipUnless(has_itk, "itk not installed")
class TestLoadSaveNifti(unittest.TestCase):
    def setUp(self):
        self.test_dir = tempfile.mkdtemp()

    def tearDown(self):
        shutil.rmtree(self.test_dir, ignore_errors=True)

    def nifti_rw(self, test_data, reader, writer, dtype, resample=True):
        test_data = test_data.astype(dtype)
        ndim = len(test_data.shape) - 1
        for p in TEST_NDARRAYS:
            output_ext = ".nii.gz"
            filepath = f"testfile_{ndim}d"
            saver = SaveImage(
                output_dir=self.test_dir,
                output_ext=output_ext,
                output_dtype=None,
                resample=resample,
                separate_folder=False,
                writer=writer,
            )
            meta_dict = {
                "filename_or_obj": f"{filepath}.png",
                "affine": np.eye(4),
                "original_affine": np.array([[0, 1, 0, 0], [1, 0, 0, 0], [0, 0, 1, 0], [0, 0, 0, 1]]),
            }
            test_data = MetaTensor(p(test_data), meta=meta_dict)
            self.assertEqual(test_data.meta[MetaKeys.SPACE], "RAS")
            saver(test_data)
            saved_path = os.path.join(self.test_dir, filepath + "_trans" + output_ext)
            self.assertTrue(os.path.exists(saved_path))
            loader = LoadImage(image_only=True, reader=reader, squeeze_non_spatial_dims=True, dtype=None)
            data = loader(saved_path)
            self.assertIn(dtype.__name__, str(data.dtype))
            meta = data.meta
            if meta["original_channel_dim"] == -1:
                _test_data = moveaxis(test_data, 0, -1)
            else:
                _test_data = test_data[0]
            if resample:
                _test_data = moveaxis(_test_data, 0, 1)
            assert_allclose(meta["qform_code"], 1, type_test=False)
            assert_allclose(meta["sform_code"], 1, type_test=False)
            assert_allclose(data, torch.as_tensor(_test_data))

    @parameterized.expand(itertools.product([NibabelReader, ITKReader], [NibabelWriter, "ITKWriter"]))
    def test_2d(self, reader, writer):
        test_data = np.arange(48, dtype=np.uint8).reshape(1, 6, 8)
        self.nifti_rw(test_data, reader, writer, np.uint8)
        self.nifti_rw(test_data, reader, writer, np.float32)

    @parameterized.expand(itertools.product([NibabelReader, ITKReader], [NibabelWriter, ITKWriter]))
    def test_3d(self, reader, writer):
        test_data = np.arange(48, dtype=np.uint8).reshape(1, 2, 3, 8)
        self.nifti_rw(test_data, reader, writer, np.int16)
        self.nifti_rw(test_data, reader, writer, float, False)

    @parameterized.expand(itertools.product([NibabelReader, ITKReader], ["NibabelWriter", ITKWriter]))
    def test_4d(self, reader, writer):
        test_data = np.arange(48, dtype=np.uint8).reshape(2, 1, 3, 8)
        self.nifti_rw(test_data, reader, writer, np.float64)


@unittest.skipUnless(has_itk, "itk not installed")
class TestLoadSavePNG(unittest.TestCase):
    def setUp(self):
        self.test_dir = tempfile.mkdtemp()

    def tearDown(self):
        shutil.rmtree(self.test_dir, ignore_errors=True)

    def png_rw(self, test_data, reader, writer, dtype, resample=True):
        test_data = test_data.astype(dtype)
        ndim = len(test_data.shape) - 1
        for p in TEST_NDARRAYS:
            output_ext = ".png"
            filepath = f"testfile_{ndim}d"
            saver = SaveImage(
                output_dir=self.test_dir, output_ext=output_ext, resample=resample, separate_folder=False, writer=writer
            )
            test_data = MetaTensor(p(test_data), meta={"filename_or_obj": f"{filepath}.png", "spatial_shape": (6, 8)})
            saver(test_data)
            saved_path = os.path.join(self.test_dir, filepath + "_trans" + output_ext)
            self.assertTrue(os.path.exists(saved_path))
            loader = LoadImage(image_only=True, reader=reader)
            data = loader(saved_path)
            meta = data.meta
            if meta["original_channel_dim"] == -1:
                _test_data = moveaxis(test_data, 0, -1)
            else:
                _test_data = test_data[0]
            assert_allclose(data, torch.as_tensor(_test_data))

    @parameterized.expand(itertools.product([PILReader, ITKReader], [PILWriter, ITKWriter]))
    def test_2d(self, reader, writer):
        test_data = np.arange(48, dtype=np.uint8).reshape(1, 6, 8)
        self.png_rw(test_data, reader, writer, np.uint8)

    @parameterized.expand(itertools.product([PILReader, ITKReader], ["monai.data.PILWriter", ITKWriter]))
    def test_rgb(self, reader, writer):
        test_data = np.arange(48, dtype=np.uint8).reshape(3, 2, 8)
        self.png_rw(test_data, reader, writer, np.uint8, False)


class TestRegRes(unittest.TestCase):
    def test_0_default(self):
        self.assertTrue(len(resolve_writer(".png")) > 0, "has png writer")
        self.assertTrue(len(resolve_writer(".nrrd")) > 0, "has nrrd writer")
        self.assertTrue(len(resolve_writer("unknown")) > 0, "has writer")
        register_writer("unknown1", lambda: (_ for _ in ()).throw(OptionalImportError))
        with self.assertRaises(OptionalImportError):
            resolve_writer("unknown1")

    def test_1_new(self):
        register_writer("new", lambda x: x + 1)
        register_writer("new2", lambda x: x + 1)
        self.assertEqual(resolve_writer("new")[0](0), 1)


@unittest.skipUnless(has_itk, "itk not installed")
class TestLoadSaveNrrd(unittest.TestCase):
    def setUp(self):
        self.test_dir = tempfile.mkdtemp()

    def tearDown(self):
        shutil.rmtree(self.test_dir, ignore_errors=True)

    def nrrd_rw(self, test_data, reader, writer, dtype, resample=True):
        test_data = test_data.astype(dtype)
        ndim = len(test_data.shape)
        for p in TEST_NDARRAYS:
            output_ext = ".nrrd"
            filepath = f"testfile_{ndim}d"
            saver = SaveImage(
                output_dir=self.test_dir, output_ext=output_ext, resample=resample, separate_folder=False, writer=writer
            ).set_options(init_kwargs={"affine_lps_to_ras": True})
            test_data = MetaTensor(
                p(test_data), meta={"filename_or_obj": f"{filepath}{output_ext}", "spatial_shape": test_data.shape}
            )
            saver(test_data)
            saved_path = os.path.join(self.test_dir, filepath + "_trans" + output_ext)
            loader = LoadImage(image_only=True, reader=reader)
            data = loader(saved_path)
            assert_allclose(data, torch.as_tensor(test_data))

    @parameterized.expand(itertools.product([NrrdReader, ITKReader], [ITKWriter, ITKWriter]))
    def test_2d(self, reader, writer):
        test_data = np.random.randn(8, 8).astype(np.float32)
        self.nrrd_rw(test_data, reader, writer, np.float32)

    @parameterized.expand(itertools.product([NrrdReader, ITKReader], [ITKWriter, ITKWriter]))
    def test_3d(self, reader, writer):
        test_data = np.random.randn(8, 8, 8).astype(np.float32)
        self.nrrd_rw(test_data, reader, writer, np.float32)


if __name__ == "__main__":
    unittest.main()



# Copyright (c) MONAI Consortium
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#     http://www.apache.org/licenses/LICENSE-2.0
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import annotations

import math

from torch.optim import Optimizer
from torch.optim.lr_scheduler import LambdaLR, _LRScheduler

__all__ = ["LinearLR", "ExponentialLR"]


class _LRSchedulerMONAI(_LRScheduler):
    """Base class for increasing the learning rate between two boundaries over a number
    of iterations"""

    def __init__(self, optimizer: Optimizer, end_lr: float, num_iter: int, last_epoch: int = -1) -> None:
        """
        Args:
            optimizer: wrapped optimizer.
            end_lr: the final learning rate.
            num_iter: the number of iterations over which the test occurs.
            last_epoch: the index of last epoch.
        Returns:
            None
        """
        self.end_lr = end_lr
        self.num_iter = num_iter
        super().__init__(optimizer, last_epoch)


class LinearLR(_LRSchedulerMONAI):
    """Linearly increases the learning rate between two boundaries over a number of
    iterations.
    """

    def get_lr(self):
        r = self.last_epoch / (self.num_iter - 1)
        return [base_lr + r * (self.end_lr - base_lr) for base_lr in self.base_lrs]


class ExponentialLR(_LRSchedulerMONAI):
    """Exponentially increases the learning rate between two boundaries over a number of
    iterations.
    """

    def get_lr(self):
        r = self.last_epoch / (self.num_iter - 1)
        return [base_lr * (self.end_lr / base_lr) ** r for base_lr in self.base_lrs]


class WarmupCosineSchedule(LambdaLR):
    """Linear warmup and then cosine decay.
    Based on https://huggingface.co/ implementation.
    """

    def __init__(
        self,
        optimizer: Optimizer,
        warmup_steps: int,
        t_total: int,
        cycles: float = 0.5,
        last_epoch: int = -1,
        warmup_multiplier: float = 0,
    ) -> None:
        """
        Args:
            optimizer: wrapped optimizer.
            warmup_steps: number of warmup iterations.
            t_total: total number of training iterations.
            cycles: cosine cycles parameter.
            last_epoch: the index of last epoch.
            warmup_multiplier: if provided, starts the linear warmup from this fraction of the initial lr.
                Must be in 0..1 interval. Defaults to 0
        Returns:
            None
        """
        self.warmup_steps = min(max(warmup_steps, 0), t_total)
        self.warmup_multiplier = warmup_multiplier
        self.t_total = t_total
        self.cycles = cycles
        if warmup_multiplier < 0 or warmup_multiplier > 1:
            raise ValueError("warmup_multiplier must be in 0..1 range")
        super().__init__(optimizer, self.lr_lambda, last_epoch)

    def lr_lambda(self, step):
        if step < self.warmup_steps:
            f = float(step) / float(max(1.0, self.warmup_steps))
            return self.warmup_multiplier + (1 - self.warmup_multiplier) * f
        progress = float(step - self.warmup_steps) / float(max(1, self.t_total - self.warmup_steps))
        return max(0.0, 0.5 * (1.0 + math.cos(math.pi * float(self.cycles) * 2.0 * progress)))



# Copyright (c) MONAI Consortium
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#     http://www.apache.org/licenses/LICENSE-2.0
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import annotations

import os
import tempfile
import unittest

import torch
from parameterized import parameterized

from monai.data.meta_tensor import MetaTensor
from monai.transforms import SaveImage
from monai.utils import optional_import

_, has_itk = optional_import("itk", allow_namespace_pkg=True)

TEST_CASE_1 = [torch.randint(0, 255, (1, 2, 3, 4)), {"filename_or_obj": "testfile0.nii.gz"}, ".nii.gz", False]

TEST_CASE_2 = [torch.randint(0, 255, (1, 2, 3, 4)), None, ".nii.gz", False]

TEST_CASE_3 = [torch.randint(0, 255, (1, 2, 3, 4)), {"filename_or_obj": "testfile0.nrrd"}, ".nrrd", False]

TEST_CASE_4 = [
    torch.randint(0, 255, (3, 2, 4, 5), dtype=torch.uint8),
    {"filename_or_obj": "testfile0.dcm"},
    ".dcm",
    False,
]


@unittest.skipUnless(has_itk, "itk not installed")
class TestSaveImage(unittest.TestCase):
    @parameterized.expand([TEST_CASE_1, TEST_CASE_2, TEST_CASE_3, TEST_CASE_4])
    def test_saved_content(self, test_data, meta_data, output_ext, resample):
        if meta_data is not None:
            test_data = MetaTensor(test_data, meta=meta_data)

        with tempfile.TemporaryDirectory() as tempdir:
            trans = SaveImage(
                output_dir=tempdir,
                output_ext=output_ext,
                resample=resample,
                separate_folder=False,  # test saving into the same folder
                output_name_formatter=lambda x, xform: dict(subject=x["filename_or_obj"] if x else "0"),
            )
            trans(test_data)

            filepath = "testfile0" if meta_data is not None else "0"
            self.assertTrue(os.path.exists(os.path.join(tempdir, filepath + "_trans" + output_ext)))


if __name__ == "__main__":
    unittest.main()



# Copyright (c) MONAI Consortium
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#     http://www.apache.org/licenses/LICENSE-2.0
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import annotations

import unittest

import numpy as np
from parameterized import parameterized

from monai.transforms import GridDistortiond
from tests.utils import TEST_NDARRAYS_ALL, assert_allclose

TESTS = []
num_cells = (2, 2)
distort_steps = [(1.5,) * (1 + n_c) for n_c in num_cells]
for p in TEST_NDARRAYS_ALL:
    img = np.indices([6, 6]).astype(np.float32)
    TESTS.append(
        [
            dict(
                keys=["img", "mask"],
                num_cells=num_cells,
                distort_steps=distort_steps,
                mode=["bilinear", "nearest"],
                padding_mode=["reflection", "zeros"],
            ),
            {"img": p(img), "mask": p(np.ones_like(img[:1]))},
            p(
                np.array(
                    [
                        [
                            [0.0, 0.0, 0.0, 0.0, 0.0, 0.0],
                            [2.25, 2.25, 2.25, 2.25, 2.25, 2.25],
                            [4.5, 4.5, 4.5, 4.5, 4.5, 4.5],
                            [4.5, 4.5, 4.5, 4.5, 4.5, 4.5],
                            [4.2500, 4.2500, 4.2500, 4.2500, 4.2500, 4.2500],
                            [2.0000, 2.0000, 2.0000, 2.0000, 2.0000, 2.0000],
                        ],
                        [
                            [0.0000, 2.2500, 4.5000, 4.5000, 4.2500, 2.0000],
                            [0.0000, 2.2500, 4.5000, 4.5000, 4.2500, 2.0000],
                            [0.0000, 2.2500, 4.5000, 4.5000, 4.2500, 2.0000],
                            [0.0000, 2.2500, 4.5000, 4.5000, 4.2500, 2.0000],
                            [0.0000, 2.2500, 4.5000, 4.5000, 4.2500, 2.0000],
                            [0.0000, 2.2500, 4.5000, 4.5000, 4.2500, 2.0000],
                        ],
                    ]
                ).astype(np.float32)
            ),
            p(
                np.array(
                    [
                        [
                            [1.0, 1.0, 1.0, 1.0, 0.0, 0.0],
                            [1.0, 1.0, 1.0, 1.0, 0.0, 0.0],
                            [1.0, 1.0, 1.0, 1.0, 0.0, 0.0],
                            [1.0, 1.0, 1.0, 1.0, 0.0, 0.0],
                            [0.0, 0.0, 0.0, 0.0, 0.0, 0.0],
                            [0.0, 0.0, 0.0, 0.0, 0.0, 0.0],
                        ]
                    ]
                ).astype(np.float32)
            ),
        ]
    )


class TestGridDistortiond(unittest.TestCase):
    @parameterized.expand(TESTS)
    def test_grid_distortiond(self, input_param, input_data, expected_val_img, expected_val_mask):
        g = GridDistortiond(**input_param)
        result = g(input_data)
        assert_allclose(result["mask"], expected_val_mask, type_test=False, rtol=1e-4, atol=1e-4)
        assert_allclose(result["img"].shape, expected_val_img.shape, type_test=False, rtol=1e-4, atol=1e-4)


if __name__ == "__main__":
    unittest.main()



# Copyright (c) MONAI Consortium
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#     http://www.apache.org/licenses/LICENSE-2.0
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import annotations

import unittest

import numpy as np
from parameterized import parameterized

from monai.transforms import CuCIMd
from monai.utils import optional_import, set_determinism
from tests.utils import HAS_CUPY, skip_if_no_cuda

_, has_cut = optional_import("cucim.core.operations.expose.transform")
cp, _ = optional_import("cupy")

set_determinism(seed=0)

TEST_CASE_COLOR_JITTER_1 = [
    {"name": "color_jitter", "brightness": 0.0, "contrast": 0.0, "saturation": 0.0, "hue": 0.0},
    np.array([[[0.0, 1.0], [1.0, 2.0]], [[0.0, 1.0], [1.0, 2.0]], [[0.0, 1.0], [1.0, 2.0]]], dtype=np.float32),
    np.array([[[0.0, 1.0], [1.0, 2.0]], [[0.0, 1.0], [1.0, 2.0]], [[0.0, 1.0], [1.0, 2.0]]], dtype=np.float32),
]

TEST_CASE_COLOR_JITTER_2 = [
    {"name": "color_jitter", "brightness": 0.0, "contrast": 0.0, "saturation": 0.0, "hue": 0.0},
    np.array([[[0, 1], [2, 3]], [[0, 1], [2, 3]], [[0, 1], [2, 3]]], dtype=np.uint8),
    np.array([[[0, 1], [2, 3]], [[0, 1], [2, 3]], [[0, 1], [2, 3]]], dtype=np.uint8),
]

TEST_CASE_FLIP_1 = [
    {"name": "image_flip", "spatial_axis": -1},
    np.array([[[0.0, 1.0], [2.0, 3.0]], [[0.0, 1.0], [2.0, 3.0]], [[0.0, 1.0], [2.0, 3.0]]], dtype=np.float32),
    np.array([[[1.0, 0.0], [3.0, 2.0]], [[1.0, 0.0], [3.0, 2.0]], [[1.0, 0.0], [3.0, 2.0]]], dtype=np.float32),
]

TEST_CASE_ROTATE_1 = [
    {"name": "image_rotate_90", "k": 1, "spatial_axis": (-2, -1)},
    np.array([[[0.0, 1.0], [2.0, 3.0]], [[0.0, 1.0], [2.0, 3.0]], [[0.0, 1.0], [2.0, 3.0]]], dtype=np.float32),
    np.array([[[1.0, 3.0], [0.0, 2.0]], [[1.0, 3.0], [0.0, 2.0]], [[1.0, 3.0], [0.0, 2.0]]], dtype=np.float32),
]

TEST_CASE_SCALE_INTENSITY_1 = [
    {"name": "scale_intensity_range", "a_min": 0.0, "a_max": 4.0, "b_min": 0.0, "b_max": 1.0, "clip": False},
    np.array([[[0.0, 1.0], [2.0, 3.0]], [[0.0, 1.0], [2.0, 3.0]], [[0.0, 1.0], [2.0, 3.0]]], dtype=np.float32),
    np.array([[[0.0, 0.25], [0.5, 0.75]], [[0.0, 0.25], [0.5, 0.75]], [[0.0, 0.25], [0.5, 0.75]]], dtype=np.float32),
]

TEST_CASE_ZOOM_1 = [
    {"name": "zoom", "zoom_factor": (0.5, 0.5)},
    np.mgrid[:3, 1:4].astype(dtype=np.float32),
    np.concatenate([np.ones((1, 3, 3), dtype=np.float32) * 1.0, np.ones((1, 3, 3), dtype=np.float32) * 2.0]),
]


@skip_if_no_cuda
@unittest.skipUnless(HAS_CUPY, "CuPy is required.")
@unittest.skipUnless(has_cut, "cuCIM transforms are required.")
class TestCuCIMDict(unittest.TestCase):
    @parameterized.expand(
        [
            TEST_CASE_COLOR_JITTER_1,
            TEST_CASE_COLOR_JITTER_2,
            TEST_CASE_FLIP_1,
            TEST_CASE_ROTATE_1,
            TEST_CASE_SCALE_INTENSITY_1,
            TEST_CASE_ZOOM_1,
        ]
    )
    def test_tramsforms_numpy_single(self, params, input, expected):
        input = {"image": input}
        output = CuCIMd(keys="image", **params)(input)["image"]
        self.assertTrue(output.dtype == expected.dtype)
        self.assertTrue(isinstance(output, np.ndarray))
        cp.testing.assert_allclose(output, expected)

    @parameterized.expand(
        [
            TEST_CASE_COLOR_JITTER_1,
            TEST_CASE_COLOR_JITTER_2,
            TEST_CASE_FLIP_1,
            TEST_CASE_ROTATE_1,
            TEST_CASE_SCALE_INTENSITY_1,
            TEST_CASE_ZOOM_1,
        ]
    )
    def test_tramsforms_numpy_batch(self, params, input, expected):
        input = {"image": input[cp.newaxis, ...]}
        expected = expected[cp.newaxis, ...]
        output = CuCIMd(keys="image", **params)(input)["image"]
        self.assertTrue(output.dtype == expected.dtype)
        self.assertTrue(isinstance(output, np.ndarray))
        cp.testing.assert_allclose(output, expected)

    @parameterized.expand(
        [
            TEST_CASE_COLOR_JITTER_1,
            TEST_CASE_COLOR_JITTER_2,
            TEST_CASE_FLIP_1,
            TEST_CASE_ROTATE_1,
            TEST_CASE_SCALE_INTENSITY_1,
            TEST_CASE_ZOOM_1,
        ]
    )
    def test_tramsforms_cupy_single(self, params, input, expected):
        input = {"image": cp.asarray(input)}
        expected = cp.asarray(expected)
        output = CuCIMd(keys="image", **params)(input)["image"]
        self.assertTrue(output.dtype == expected.dtype)
        self.assertTrue(isinstance(output, cp.ndarray))
        cp.testing.assert_allclose(output, expected)

    @parameterized.expand(
        [
            TEST_CASE_COLOR_JITTER_1,
            TEST_CASE_COLOR_JITTER_2,
            TEST_CASE_FLIP_1,
            TEST_CASE_ROTATE_1,
            TEST_CASE_SCALE_INTENSITY_1,
            TEST_CASE_ZOOM_1,
        ]
    )
    def test_tramsforms_cupy_batch(self, params, input, expected):
        input = {"image": cp.asarray(input)[cp.newaxis, ...]}
        expected = cp.asarray(expected)[cp.newaxis, ...]
        output = CuCIMd(keys="image", **params)(input)["image"]
        self.assertTrue(output.dtype == expected.dtype)
        self.assertTrue(isinstance(output, cp.ndarray))
        cp.testing.assert_allclose(output, expected)


if __name__ == "__main__":
    unittest.main()



# Copyright (c) MONAI Consortium
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#     http://www.apache.org/licenses/LICENSE-2.0
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import annotations

import shutil
import tempfile
import unittest

import numpy as np

from monai.data import LMDBDataset, json_hashing
from monai.transforms import Transform
from tests.utils import DistCall, DistTestCase, skip_if_windows


class _InplaceXform(Transform):
    def __call__(self, data):
        if data:
            data[0] = data[0] + np.pi
        else:
            data.append(1)
        return data


@skip_if_windows
class TestMPLMDBDataset(DistTestCase):
    def setUp(self):
        self.tempdir = tempfile.mkdtemp()

    def tearDown(self):
        shutil.rmtree(self.tempdir)

    @DistCall(nnodes=1, nproc_per_node=1)
    def test_mp_cache(self):
        items = [[list(range(i))] for i in range(5)]

        ds = LMDBDataset(items, transform=_InplaceXform(), cache_dir=self.tempdir, lmdb_kwargs={"map_size": 10 * 1024})
        self.assertEqual(items, [[[]], [[0]], [[0, 1]], [[0, 1, 2]], [[0, 1, 2, 3]]])
        ds1 = LMDBDataset(items, transform=_InplaceXform(), cache_dir=self.tempdir, lmdb_kwargs={"map_size": 10 * 1024})
        self.assertEqual(list(ds1), list(ds))
        self.assertEqual(items, [[[]], [[0]], [[0, 1]], [[0, 1, 2]], [[0, 1, 2, 3]]])

        ds = LMDBDataset(
            items,
            transform=_InplaceXform(),
            cache_dir=self.tempdir,
            lmdb_kwargs={"map_size": 10 * 1024},
            hash_func=json_hashing,
        )
        self.assertEqual(items, [[[]], [[0]], [[0, 1]], [[0, 1, 2]], [[0, 1, 2, 3]]])
        ds1 = LMDBDataset(
            items,
            transform=_InplaceXform(),
            cache_dir=self.tempdir,
            lmdb_kwargs={"map_size": 10 * 1024},
            hash_func=json_hashing,
        )
        self.assertEqual(list(ds1), list(ds))
        self.assertEqual(items, [[[]], [[0]], [[0, 1]], [[0, 1, 2]], [[0, 1, 2, 3]]])

        self.assertTrue(isinstance(ds1.info(), dict))


if __name__ == "__main__":
    unittest.main()



# Copyright (c) MONAI Consortium
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#     http://www.apache.org/licenses/LICENSE-2.0
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import annotations

import unittest
from typing import Any, Sequence

import torch
from parameterized import parameterized

from monai.networks import eval_mode
from monai.networks.nets import DynUNet
from monai.utils import optional_import
from tests.utils import assert_allclose, skip_if_no_cuda, skip_if_windows, test_script_save

InstanceNorm3dNVFuser, _ = optional_import("apex.normalization", name="InstanceNorm3dNVFuser")

device = "cuda" if torch.cuda.is_available() else "cpu"

strides: Sequence[Sequence[int] | int]
kernel_size: Sequence[Any]
expected_shape: Sequence[Any]

TEST_CASE_DYNUNET_2D = []
out_channels = 2
in_size = 64
spatial_dims = 2
for kernel_size in [(3, 3, 3, 1), ((3, 1), 1, (3, 3), (1, 1))]:
    for strides in [(1, 1, 1, 1), (2, 2, 2, 1)]:
        expected_shape = (1, out_channels, *[in_size // strides[0]] * spatial_dims)
        for in_channels in [2, 3]:
            for res_block in [True, False]:
                test_case = [
                    {
                        "spatial_dims": spatial_dims,
                        "in_channels": in_channels,
                        "out_channels": out_channels,
                        "kernel_size": kernel_size,
                        "strides": strides,
                        "upsample_kernel_size": strides[1:],
                        "norm_name": "batch",
                        "act_name": ("leakyrelu", {"inplace": True, "negative_slope": 0.2}),
                        "deep_supervision": False,
                        "res_block": res_block,
                        "dropout": None,
                    },
                    (1, in_channels, in_size, in_size),
                    expected_shape,
                ]
                TEST_CASE_DYNUNET_2D.append(test_case)

TEST_CASE_DYNUNET_3D = []  # in 3d cases, also test anisotropic kernel/strides
in_channels = 1
in_size = 64
for out_channels in [2, 3]:
    expected_shape = (1, out_channels, 64, 32, 64)
    for res_block in [True, False]:
        test_case = [
            {
                "spatial_dims": 3,
                "in_channels": in_channels,
                "out_channels": out_channels,
                "kernel_size": (3, (1, 1, 3), 3, 3),
                "strides": ((1, 2, 1), 2, 2, 1),
                "upsample_kernel_size": (2, 2, 1),
                "filters": (64, 96, 128, 192),
                "norm_name": ("INSTANCE", {"affine": True}),
                "deep_supervision": True,
                "res_block": res_block,
                "dropout": ("alphadropout", {"p": 0.25}),
            },
            (1, in_channels, in_size, in_size, in_size),
            expected_shape,
        ]
        TEST_CASE_DYNUNET_3D.append(test_case)

TEST_CASE_DEEP_SUPERVISION = []
for spatial_dims in [2, 3]:
    for res_block in [True, False]:
        for deep_supr_num in [1, 2]:
            for strides in [(1, 2, 1, 2, 1), (2, 2, 2, 1), (2, 1, 1, 2, 2)]:
                scale = strides[0]
                test_case = [
                    {
                        "spatial_dims": spatial_dims,
                        "in_channels": 1,
                        "out_channels": 2,
                        "kernel_size": [3] * len(strides),
                        "strides": strides,
                        "upsample_kernel_size": strides[1:],
                        "norm_name": ("group", {"num_groups": 16}),
                        "deep_supervision": True,
                        "deep_supr_num": deep_supr_num,
                        "res_block": res_block,
                    },
                    (1, 1, *[in_size] * spatial_dims),
                    (1, 1 + deep_supr_num, 2, *[in_size // scale] * spatial_dims),
                ]
                TEST_CASE_DEEP_SUPERVISION.append(test_case)


class TestDynUNet(unittest.TestCase):
    @parameterized.expand(TEST_CASE_DYNUNET_3D)
    def test_shape(self, input_param, input_shape, expected_shape):
        net = DynUNet(**input_param).to(device)
        if "alphadropout" in input_param.get("dropout"):
            self.assertTrue(any(isinstance(x, torch.nn.AlphaDropout) for x in net.modules()))
        with eval_mode(net):
            result = net(torch.randn(input_shape).to(device))
            self.assertEqual(result.shape, expected_shape)

    def test_script(self):
        input_param, input_shape, _ = TEST_CASE_DYNUNET_2D[0]
        net = DynUNet(**input_param)
        test_data = torch.randn(input_shape)
        test_script_save(net, test_data)


@skip_if_no_cuda
@skip_if_windows
class TestDynUNetWithInstanceNorm3dNVFuser(unittest.TestCase):
    def setUp(self):
        try:
            layer = InstanceNorm3dNVFuser(num_features=1, affine=False).to("cuda:0")
            inp = torch.randn([1, 1, 1, 1, 1]).to("cuda:0")
            out = layer(inp)
            del inp, out, layer
        except Exception:
            self.skipTest("NVFuser not available")

    @parameterized.expand([TEST_CASE_DYNUNET_3D[0]])
    def test_consistency(self, input_param, input_shape, _):
        for eps in [1e-4, 1e-5]:
            for momentum in [0.1, 0.01]:
                for affine in [True, False]:
                    norm_param = {"eps": eps, "momentum": momentum, "affine": affine}
                    input_param["norm_name"] = ("instance", norm_param)
                    input_param_fuser = input_param.copy()
                    input_param_fuser["norm_name"] = ("instance_nvfuser", norm_param)
                    for memory_format in [torch.contiguous_format, torch.channels_last_3d]:
                        net = DynUNet(**input_param).to("cuda:0", memory_format=memory_format)
                        net_fuser = DynUNet(**input_param_fuser).to("cuda:0", memory_format=memory_format)
                        net_fuser.load_state_dict(net.state_dict())

                        input_tensor = torch.randn(input_shape).to("cuda:0", memory_format=memory_format)
                        with eval_mode(net):
                            result = net(input_tensor)
                        with eval_mode(net_fuser):
                            result_fuser = net_fuser(input_tensor)

                        assert_allclose(result, result_fuser, rtol=1e-4, atol=1e-4)


class TestDynUNetDeepSupervision(unittest.TestCase):
    @parameterized.expand(TEST_CASE_DEEP_SUPERVISION)
    def test_shape(self, input_param, input_shape, expected_shape):
        net = DynUNet(**input_param).to(device)
        with torch.no_grad():
            results = net(torch.randn(input_shape).to(device))
            self.assertEqual(results.shape, expected_shape)


if __name__ == "__main__":
    unittest.main()



# Copyright (c) MONAI Consortium
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#     http://www.apache.org/licenses/LICENSE-2.0
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import annotations

import unittest

import torch
from parameterized import parameterized

from monai.networks import eval_mode
from monai.networks.layers import Act, Norm
from monai.networks.nets import UNet
from tests.utils import test_script_save

device = "cuda" if torch.cuda.is_available() else "cpu"

TEST_CASE_0 = [  # single channel 2D, batch 16, no residual
    {
        "spatial_dims": 2,
        "in_channels": 1,
        "out_channels": 3,
        "channels": (16, 32, 64),
        "strides": (2, 2),
        "num_res_units": 0,
    },
    (16, 1, 32, 32),
    (16, 3, 32, 32),
]

TEST_CASE_1 = [  # single channel 2D, batch 16
    {
        "spatial_dims": 2,
        "in_channels": 1,
        "out_channels": 3,
        "channels": (16, 32, 64),
        "strides": (2, 2),
        "num_res_units": 1,
    },
    (16, 1, 32, 32),
    (16, 3, 32, 32),
]

TEST_CASE_2 = [  # single channel 3D, batch 16
    {
        "spatial_dims": 3,
        "in_channels": 1,
        "out_channels": 3,
        "channels": (16, 32, 64),
        "strides": (2, 2),
        "num_res_units": 1,
    },
    (16, 1, 32, 24, 48),
    (16, 3, 32, 24, 48),
]

TEST_CASE_3 = [  # 4-channel 3D, batch 16
    {
        "spatial_dims": 3,
        "in_channels": 4,
        "out_channels": 3,
        "channels": (16, 32, 64),
        "strides": (2, 2),
        "num_res_units": 1,
    },
    (16, 4, 32, 64, 48),
    (16, 3, 32, 64, 48),
]

TEST_CASE_4 = [  # 4-channel 3D, batch 16, batch normalization
    {
        "spatial_dims": 3,
        "in_channels": 4,
        "out_channels": 3,
        "channels": (16, 32, 64),
        "strides": (2, 2),
        "num_res_units": 1,
        "norm": Norm.BATCH,
    },
    (16, 4, 32, 64, 48),
    (16, 3, 32, 64, 48),
]

TEST_CASE_5 = [  # 4-channel 3D, batch 16, LeakyReLU activation
    {
        "spatial_dims": 3,
        "in_channels": 4,
        "out_channels": 3,
        "channels": (16, 32, 64),
        "strides": (2, 2),
        "num_res_units": 1,
        "act": (Act.LEAKYRELU, {"negative_slope": 0.2}),
        "adn_ordering": "NA",
    },
    (16, 4, 32, 64, 48),
    (16, 3, 32, 64, 48),
]

TEST_CASE_6 = [  # 4-channel 3D, batch 16, LeakyReLU activation explicit
    {
        "spatial_dims": 3,
        "in_channels": 4,
        "out_channels": 3,
        "channels": (16, 32, 64),
        "strides": (2, 2),
        "num_res_units": 1,
        "act": (torch.nn.LeakyReLU, {"negative_slope": 0.2}),
    },
    (16, 4, 32, 64, 48),
    (16, 3, 32, 64, 48),
]

CASES = [TEST_CASE_0, TEST_CASE_1, TEST_CASE_2, TEST_CASE_3, TEST_CASE_4, TEST_CASE_5, TEST_CASE_6]

ILL_CASES = [
    [
        {  # len(channels) < 2
            "spatial_dims": 2,
            "in_channels": 1,
            "out_channels": 3,
            "channels": (16,),
            "strides": (2, 2),
            "num_res_units": 0,
        }
    ],
    [
        {  # len(strides) < len(channels) - 1
            "spatial_dims": 2,
            "in_channels": 1,
            "out_channels": 3,
            "channels": (8, 8, 8),
            "strides": (2,),
            "num_res_units": 0,
        }
    ],
    [
        {  # len(kernel_size) = 3, spatial_dims = 2
            "spatial_dims": 2,
            "in_channels": 1,
            "out_channels": 3,
            "channels": (8, 8, 8),
            "strides": (2, 2),
            "kernel_size": (3, 3, 3),
        }
    ],
    [
        {  # len(up_kernel_size) = 2, spatial_dims = 3
            "spatial_dims": 3,
            "in_channels": 1,
            "out_channels": 3,
            "channels": (8, 8, 8),
            "strides": (2, 2),
            "up_kernel_size": (3, 3),
        }
    ],
]


class TestUNET(unittest.TestCase):
    @parameterized.expand(CASES)
    def test_shape(self, input_param, input_shape, expected_shape):
        net = UNet(**input_param).to(device)
        with eval_mode(net):
            result = net.forward(torch.randn(input_shape).to(device))
            self.assertEqual(result.shape, expected_shape)

    def test_script(self):
        net = UNet(
            spatial_dims=2, in_channels=1, out_channels=3, channels=(16, 32, 64), strides=(2, 2), num_res_units=0
        )
        test_data = torch.randn(16, 1, 32, 32)
        test_script_save(net, test_data)

    def test_script_without_running_stats(self):
        net = UNet(
            spatial_dims=2,
            in_channels=1,
            out_channels=3,
            channels=(16, 32, 64),
            strides=(2, 2),
            num_res_units=0,
            norm=("batch", {"track_running_stats": False}),
        )
        test_data = torch.randn(16, 1, 16, 4)
        test_script_save(net, test_data)

    def test_ill_input_shape(self):
        net = UNet(spatial_dims=2, in_channels=1, out_channels=3, channels=(16, 32, 64), strides=(2, 2))
        with eval_mode(net):
            with self.assertRaisesRegex(RuntimeError, "Sizes of tensors must match"):
                net.forward(torch.randn(2, 1, 16, 5))

    @parameterized.expand(ILL_CASES)
    def test_ill_input_hyper_params(self, input_param):
        with self.assertRaises(ValueError):
            _ = UNet(**input_param)


if __name__ == "__main__":
    unittest.main()



# Copyright (c) MONAI Consortium
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#     http://www.apache.org/licenses/LICENSE-2.0
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import annotations

import unittest
import warnings

import torch

from monai.engines import create_multigpu_supervised_trainer
from tests.utils import skip_if_no_cuda


def fake_loss(y_pred, y):
    return (y_pred[0] + y).sum()


def fake_data_stream():
    while True:
        yield torch.rand((10, 1, 64, 64)), torch.rand((10, 1, 64, 64))


class TestParallelExecution(unittest.TestCase):
    """
    Tests single GPU, multi GPU, and CPU execution with the Ignite supervised trainer.
    """

    @skip_if_no_cuda
    def test_single_gpu(self):
        device = torch.device("cuda:0")
        net = torch.nn.Conv2d(1, 1, 3, padding=1).to(device)
        opt = torch.optim.Adam(net.parameters(), 1e-3)
        trainer = create_multigpu_supervised_trainer(net, opt, fake_loss, [device])
        trainer.run(fake_data_stream(), 2, 2)

    @skip_if_no_cuda
    def test_multi_gpu(self):
        device = torch.device("cuda")
        net = torch.nn.Conv2d(1, 1, 3, padding=1).to(device)
        opt = torch.optim.Adam(net.parameters(), 1e-3)

        with warnings.catch_warnings():
            warnings.simplefilter("ignore")  # ignore warnings about imbalanced GPU memory

            trainer = create_multigpu_supervised_trainer(net, opt, fake_loss, None)

        trainer.run(fake_data_stream(), 2, 2)

    def test_cpu(self):
        net = torch.nn.Conv2d(1, 1, 3, padding=1)
        opt = torch.optim.Adam(net.parameters(), 1e-3)
        trainer = create_multigpu_supervised_trainer(net, opt, fake_loss, [])
        trainer.run(fake_data_stream(), 2, 2)


if __name__ == "__main__":
    unittest.main()



# Copyright (c) MONAI Consortium
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#     http://www.apache.org/licenses/LICENSE-2.0
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import annotations

import os
import tempfile
import unittest
from pathlib import Path

import nibabel as nib
import numpy as np

from monai.data import check_missing_files


class TestCheckMissingFiles(unittest.TestCase):
    def test_content(self):
        test_image = nib.Nifti1Image(np.random.randint(0, 2, size=[128, 128, 128]).astype(float), np.eye(4))
        with tempfile.TemporaryDirectory() as tempdir:
            nib.save(test_image, os.path.join(tempdir, "test_image1.nii.gz"))
            nib.save(test_image, os.path.join(tempdir, "test_label1.nii.gz"))
            nib.save(test_image, os.path.join(tempdir, "test_extra1.nii.gz"))
            nib.save(test_image, os.path.join(tempdir, "test_image2.nii.gz"))

            datalist = [
                {
                    "image": os.path.join(tempdir, "test_image1.nii.gz"),
                    "label": [os.path.join(tempdir, "test_label1.nii.gz"), os.path.join(tempdir, "test_extra1.nii.gz")],
                },
                {
                    "image": Path(os.path.join(tempdir, "test_image2.nii.gz")),
                    "label": Path(os.path.join(tempdir, "test_label_missing.nii.gz")),
                },
            ]

            missings = check_missing_files(datalist=datalist, keys=["image", "label"])
            self.assertEqual(len(missings), 1)
            self.assertEqual(str(missings[0]), os.path.join(tempdir, "test_label_missing.nii.gz"))

            # test with missing key and relative path
            datalist = [{"image": "test_image1.nii.gz", "label": "test_label_missing.nii.gz"}]
            missings = check_missing_files(
                datalist=datalist, keys=["image", "label", "test"], root_dir=tempdir, allow_missing_keys=True
            )
            self.assertEqual(f"{missings[0]}", os.path.join(tempdir, "test_label_missing.nii.gz"))


if __name__ == "__main__":
    unittest.main()



# Copyright (c) MONAI Consortium
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#     http://www.apache.org/licenses/LICENSE-2.0
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import annotations

import unittest

import numpy as np
import torch
from parameterized import parameterized

from monai.transforms.utils_pytorch_numpy_unification import mode, percentile
from monai.utils import set_determinism
from tests.utils import TEST_NDARRAYS, assert_allclose, skip_if_quick

TEST_MODE = []
for p in TEST_NDARRAYS:
    TEST_MODE.append([p(np.array([1, 2, 3, 4, 4, 5])), p(4), False])
    TEST_MODE.append([p(np.array([3.1, 4.1, 4.1, 5.1])), p(4.1), False])
    TEST_MODE.append([p(np.array([3.1, 4.1, 4.1, 5.1])), p(4), True])


class TestPytorchNumpyUnification(unittest.TestCase):
    def setUp(self) -> None:
        set_determinism(0)

    def test_percentile(self):
        for size in (1, 100):
            q = np.random.randint(0, 100, size=size)
            results = []
            for idx, p in enumerate(TEST_NDARRAYS):
                dtype = [np.float32, float][idx % 2]
                arr = p(np.arange(100 * 101).reshape(1, 100, 101).astype(dtype))
                results.append(percentile(arr, q))
                assert_allclose(results[0], results[-1], type_test=False, atol=1e-4, rtol=1e-4)

    @skip_if_quick
    def test_many_elements_quantile(self):  # pytorch#64947
        for p in TEST_NDARRAYS:
            for elements in (1000, 17_000_000):
                for t in [*TEST_NDARRAYS, list]:
                    x = p(np.random.randn(elements))
                    q = percentile(x, t([10, 50]))
                    if isinstance(x, torch.Tensor):
                        self.assertIsInstance(q, torch.Tensor)
                    assert_allclose(q.shape, [2], type_test=False)

    def test_fails(self):
        for p in TEST_NDARRAYS:
            for q in (-1, 101):
                arr = p(np.arange(100 * 101).reshape(1, 100, 101).astype(np.float32))
                with self.assertRaises(ValueError):
                    percentile(arr, q)

    def test_dim(self):
        q = np.random.randint(0, 100, size=50)
        results = []
        for p in TEST_NDARRAYS:
            arr = p(np.arange(6).reshape(1, 2, 3).astype(np.float32))
            results.append(percentile(arr, q, dim=1))
            assert_allclose(results[0], results[-1], type_test=False, atol=1e-4)

    @parameterized.expand(TEST_MODE)
    def test_mode(self, array, expected, to_long):
        res = mode(array, to_long=to_long)
        assert_allclose(res, expected)


if __name__ == "__main__":
    unittest.main()



# Copyright (c) MONAI Consortium
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#     http://www.apache.org/licenses/LICENSE-2.0
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import annotations

import unittest

import torch
from parameterized import parameterized

from monai.apps.detection.networks.retinanet_network import RetinaNet, resnet_fpn_feature_extractor
from monai.networks import eval_mode
from monai.networks.nets import resnet10, resnet18, resnet34, resnet50, resnet101, resnet152, resnet200
from monai.utils import ensure_tuple, optional_import
from tests.utils import SkipIfBeforePyTorchVersion, skip_if_quick, test_onnx_save, test_script_save

_, has_torchvision = optional_import("torchvision")

device = "cuda" if torch.cuda.is_available() else "cpu"
num_anchors = 7

TEST_CASE_1 = [  # 3D, batch 3, 2 input channel
    {
        "pretrained": False,
        "spatial_dims": 3,
        "n_input_channels": 2,
        "num_classes": 3,
        "conv1_t_size": 7,
        "conv1_t_stride": (2, 2, 2),
    },
    (3, 2, 32, 64, 48),
]

TEST_CASE_2 = [  # 2D, batch 2, 1 input channel
    {
        "pretrained": False,
        "spatial_dims": 2,
        "n_input_channels": 1,
        "num_classes": 3,
        "conv1_t_size": [7, 7],
        "conv1_t_stride": [2, 2],
    },
    (2, 1, 32, 64),
]

TEST_CASE_2_A = [  # 2D, batch 2, 1 input channel, shortcut type A
    {
        "pretrained": False,
        "spatial_dims": 2,
        "n_input_channels": 1,
        "num_classes": 3,
        "shortcut_type": "A",
        "conv1_t_size": (7, 7),
        "conv1_t_stride": 2,
    },
    (2, 1, 32, 64),
]

TEST_CASE_3 = [  # 1D, batch 1, 2 input channels
    {
        "pretrained": False,
        "spatial_dims": 1,
        "n_input_channels": 2,
        "num_classes": 3,
        "conv1_t_size": [3],
        "conv1_t_stride": 1,
    },
    (1, 2, 32),
]

TEST_CASE_3_A = [  # 1D, batch 1, 2 input channels
    {"pretrained": False, "spatial_dims": 1, "n_input_channels": 2, "num_classes": 3, "shortcut_type": "A"},
    (1, 2, 32),
]

TEST_CASE_4 = [  # 2D, batch 2, 1 input channel
    {"pretrained": False, "spatial_dims": 2, "n_input_channels": 1, "num_classes": 3, "feed_forward": False},
    (2, 1, 32, 64),
]

TEST_CASES = []
for case in [TEST_CASE_1, TEST_CASE_2, TEST_CASE_3, TEST_CASE_2_A, TEST_CASE_3_A]:
    for model in [resnet10, resnet18, resnet34, resnet50, resnet101, resnet152, resnet200]:
        TEST_CASES.append([model, *case])

TEST_CASES_TS = []
for case in [TEST_CASE_1]:
    for model in [resnet10, resnet18, resnet34, resnet50, resnet101, resnet152, resnet200]:
        TEST_CASES_TS.append([model, *case])


@SkipIfBeforePyTorchVersion((1, 12))
@unittest.skipUnless(has_torchvision, "Requires torchvision")
@skip_if_quick
class TestRetinaNet(unittest.TestCase):
    @parameterized.expand(TEST_CASES)
    def test_retina_shape(self, model, input_param, input_shape):
        backbone = model(**input_param)
        feature_extractor = resnet_fpn_feature_extractor(
            backbone=backbone,
            spatial_dims=input_param["spatial_dims"],
            pretrained_backbone=input_param["pretrained"],
            trainable_backbone_layers=None,
            returned_layers=[1, 2],
        )
        net = RetinaNet(
            spatial_dims=input_param["spatial_dims"],
            num_classes=input_param["num_classes"],
            num_anchors=num_anchors,
            feature_extractor=feature_extractor,
            size_divisible=32,
        ).to(device)

        with eval_mode(net):
            result = net.forward(torch.randn(input_shape).to(device))

            base_stride = ensure_tuple(input_param["conv1_t_stride"])[0] if "conv1_t_stride" in input_param else 1
            expected_cls_channel = input_param["num_classes"] * num_anchors
            expected_cls_shape = tuple(
                (input_shape[0], expected_cls_channel)
                + tuple(input_shape[2 + a] // s // base_stride for a in range(input_param["spatial_dims"]))
                for s in [2, 4, 8]
            )
            expected_box_channel = 2 * input_param["spatial_dims"] * num_anchors
            expected_box_shape = tuple(
                (input_shape[0], expected_box_channel)
                + tuple(input_shape[2 + a] // s // base_stride for a in range(input_param["spatial_dims"]))
                for s in [2, 4, 8]
            )

            self.assertEqual(tuple(cc.shape for cc in result[net.cls_key]), expected_cls_shape)
            self.assertEqual(tuple(cc.shape for cc in result[net.box_reg_key]), expected_box_shape)

    @parameterized.expand(TEST_CASES_TS)
    def test_script(self, model, input_param, input_shape):
        try:
            idx = int(self.id().split("test_script_")[-1])
        except BaseException:
            idx = 0
        idx %= 3
        # test whether support torchscript
        data = torch.randn(input_shape)
        backbone = model(**input_param)
        if idx == 0:
            test_script_save(backbone, data)
            return
        feature_extractor = resnet_fpn_feature_extractor(
            backbone=backbone,
            spatial_dims=input_param["spatial_dims"],
            pretrained_backbone=input_param["pretrained"],
            trainable_backbone_layers=None,
            returned_layers=[1, 2],
        )
        if idx == 1:
            test_script_save(feature_extractor, data)
            return
        net = RetinaNet(
            spatial_dims=input_param["spatial_dims"],
            num_classes=input_param["num_classes"],
            num_anchors=num_anchors,
            feature_extractor=feature_extractor,
            size_divisible=32,
        )
        if idx == 2:
            test_script_save(net, data)

    @parameterized.expand(TEST_CASES_TS)
    def test_onnx(self, model, input_param, input_shape):
        try:
            idx = int(self.id().split("test_onnx_")[-1])
        except BaseException:
            idx = 0
        idx %= 3
        # test whether support torchscript
        data = torch.randn(input_shape)
        backbone = model(**input_param)
        if idx == 0:
            test_onnx_save(backbone, data, rtol=2e-3)
            return
        feature_extractor = resnet_fpn_feature_extractor(
            backbone=backbone,
            spatial_dims=input_param["spatial_dims"],
            pretrained_backbone=input_param["pretrained"],
            trainable_backbone_layers=None,
            returned_layers=[1, 2],
        )
        if idx == 1:
            test_onnx_save(feature_extractor, data, rtol=2e-3)
            return
        net = RetinaNet(
            spatial_dims=input_param["spatial_dims"],
            num_classes=input_param["num_classes"],
            num_anchors=num_anchors,
            feature_extractor=feature_extractor,
            size_divisible=32,
        )
        if idx == 2:
            test_onnx_save(net, data, rtol=2e-3)


if __name__ == "__main__":
    unittest.main()
